import logging

from pytest_zebrunner.api.client import ZebrunnerAPI
from pytest_zebrunner.api.models import LabelModel
from pytest_zebrunner.context import zebrunner_context


class AgentException(Exception):
    pass


class BaseTcm:
    @staticmethod
    def _attach_label(name: str, value: str) -> None:
        api = ZebrunnerAPI(zebrunner_context.settings.server.hostname, zebrunner_context.settings.server.access_token)
        api.auth()
        if not zebrunner_context.test_run_is_active:
            logging.error(f"Failed to attach label '{name}: {value}' to test run because it has not been started yet.")
            return
        label = LabelModel(key=name, value=value)
        api.send_labels([label], zebrunner_context.test_run_id, zebrunner_context.test_id)

    @staticmethod
    def _verify_no_tests() -> None:
        if zebrunner_context.test_is_active:
            raise AgentException(
                "The TCM configuration must be provided before start of tests. Hint: move the "
                "configuration to the code block which is executed before all tests."
            )



from pytest_zebrunner.tcm.base import BaseTcm


class Zephyr(BaseTcm):
    _SYNC_ENABLED = "com.zebrunner.app/tcm.zephyr.sync.enabled"
    _SYNC_REAL_TIME = "com.zebrunner.app/tcm.zephyr.sync.real-time"
    _TEST_CYCLE_KEY = "com.zebrunner.app/tcm.zephyr.test-cycle-key"
    _JIRA_PROJECT_KEY = "com.zebrunner.app/tcm.zephyr.jira-project-key"
    _TEST_CASE_KEY = "com.zebrunner.app/tcm.zephyr.test-case-key"

    @staticmethod
    def disable_sync() -> None:
        Zephyr._verify_no_tests()
        Zephyr._attach_label(Zephyr._SYNC_ENABLED, "false")

    @staticmethod
    def enable_real_time_sync() -> None:
        Zephyr._verify_no_tests()
        Zephyr._attach_label(Zephyr._SYNC_REAL_TIME, "true")

    @staticmethod
    def set_test_cycle_key(key: str) -> None:
        Zephyr._verify_no_tests()
        Zephyr._attach_label(Zephyr._TEST_CYCLE_KEY, key)

    @staticmethod
    def set_jira_project_key(key: str) -> None:
        Zephyr._verify_no_tests()
        Zephyr._attach_label(Zephyr._JIRA_PROJECT_KEY, key)

    @staticmethod
    def set_test_case_key(key: str) -> None:
        Zephyr._attach_label(Zephyr._TEST_CASE_KEY, key)



from datetime import datetime, timezone
from enum import Enum
from typing import Dict, List, Optional
from uuid import uuid4

from pydantic import BaseModel, Field


def generate_uuid() -> str:
    """
    Generate an universal unique identifier.

    Returns:
         (str): Universal unique identifier (uuid).
    """
    return str(uuid4())


def generate_datetime_str() -> str:
    """
    Generate a DateTime string in ISO format.

    Returns:
        (str): DateTime in ISO format.
    """
    return (datetime.utcnow()).replace(tzinfo=timezone.utc).isoformat()


class TestStatus(Enum):
    UNKNOWN = "UNKNOWN"
    IN_PROGRESS = "IN_PROGRESS"
    PASSED = "PASSED"
    FAILED = "FAILED"
    SKIPPED = "SKIPPED"
    ABORTED = "ABORTED"
    QUEUED = "QUEUED"


class NotificationsType(Enum):
    EMAIL_RECIPIENTS = "EMAIL_RECIPIENTS"
    MS_TEAMS_CHANNELS = "MS_TEAMS_CHANNELS"
    SLACK_CHANNELS = "SLACK_CHANNELS"


def to_camel_case(key: str) -> str:
    """
    Convert a string from snake case to camel case.

    Args:
        key (str): String to be converted to camel case.

    Returns:
        (str): New string in camel case.

    Example:
        this_is_snake_case.
        ThisIsCamelCase.
    """
    parts = key.split("_")
    for i, part in enumerate(parts[1:]):
        parts[i + 1] = part.capitalize()

    return "".join(parts)


class CamelModel(BaseModel):
    class Config:
        alias_generator = to_camel_case
        allow_population_by_field_name = True


class TestRunConfigModel(CamelModel):
    environment: Optional[str] = None
    build: Optional[str] = None
    treat_skips_as_failures: bool = True


class MilestoneModel(CamelModel):
    id: Optional[int]
    name: Optional[str]


class CiContextModel(CamelModel):
    ci_type: str
    env_variables: Dict[str, str]


class NotificationTargetModel(CamelModel):
    type: str
    value: str


class NotificationsModel(CamelModel):
    notify_on_each_failure: bool = False
    targets: Optional[List[NotificationTargetModel]] = None


class StartTestRunModel(CamelModel):
    name: str
    framework: str
    started_at: str = Field(default_factory=generate_datetime_str)
    uuid: str = Field(default_factory=generate_uuid)

    config: Optional[TestRunConfigModel] = None
    milestone: Optional[MilestoneModel] = None
    ci_context: Optional[CiContextModel] = None
    notifications: Optional[NotificationsModel] = None


class LabelModel(CamelModel):
    key: str
    value: str


class CorrelationDataModel(CamelModel):
    name: str


class StartTestModel(CamelModel):
    name: str
    class_name: str
    method_name: str
    uuid: str = Field(default_factory=generate_uuid)
    started_at: str = Field(default_factory=generate_datetime_str)

    correlation_data: Optional[str] = None
    maintainer: Optional[str] = None
    test_case: Optional[str] = None
    labels: Optional[List[LabelModel]] = []


class FinishTestModel(CamelModel):
    result: str
    ended_at: str = Field(default_factory=generate_datetime_str)
    reason: Optional[str] = None


class LogRecordModel(CamelModel):
    test_id: str
    level: str
    timestamp: str
    message: str


class StartTestSessionModel(CamelModel):
    session_id: str
    started_at: str = Field(default_factory=generate_datetime_str)
    desired_capabilities: dict
    capabilities: dict
    test_ids: List[int] = []


class FinishTestSessionModel(CamelModel):
    ended_at: str = Field(default_factory=generate_datetime_str)
    test_ids: List[int] = []


class AttachTestsToSessionModel(CamelModel):
    test_ids: List[int] = []


class ArtifactReferenceModel(CamelModel):
    name: str
    value: str


class TestModel(CamelModel):
    id: int
    name: str
    correlation_data: Optional[CorrelationDataModel]
    status: str
    started_at: datetime
    ended_at: datetime


class RerunDataModel(CamelModel):
    test_run_uuid: str
    run_allowed: bool
    reason: Optional[str]
    run_only_specific_tests: bool
    tests_to_run: List[TestModel]


class PlatformModel(CamelModel):
    name: str
    version: Optional[str]



import logging
from pathlib import Path
from typing import Union

from pytest_zebrunner.api.client import ZebrunnerAPI
from pytest_zebrunner.api.models import ArtifactReferenceModel, LabelModel
from pytest_zebrunner.context import zebrunner_context
from pytest_zebrunner.errors import AgentApiError, AgentError


def attach_test_screenshot(path: Union[str, Path]) -> None:
    """
    Send screenshot to zebrunner and attach it to test
    """
    if not zebrunner_context.test_is_active:
        raise AgentError("There is no active test to attach screenshot")

    try:
        api = ZebrunnerAPI(zebrunner_context.settings.server.hostname, zebrunner_context.settings.server.access_token)
        api.send_screenshot(zebrunner_context.test_run_id, zebrunner_context.test_id, path)
    except AgentApiError as e:
        logging.error("Failed to attach test screenshot", exc_info=e)


def attach_test_artifact(path: Union[str, Path]) -> None:
    """
    Send artifact to zebrunner and attach it to test. Artifact is any file from disk
    """
    if not zebrunner_context.test_is_active:
        raise AgentError("There is no active test to attach artifact")

    try:
        api = ZebrunnerAPI(zebrunner_context.settings.server.hostname, zebrunner_context.settings.server.access_token)
        api.send_artifact(path, zebrunner_context.test_run_id, zebrunner_context.test_id)
    except AgentApiError as e:
        logging.error("Failed to attach test artifact", exc_info=e)


def attach_test_run_artifact(path: Union[str, Path]) -> None:
    """
    Send artifact to zebrunner and attach it to test run. Artifact is any file from disk
    """
    if not zebrunner_context.test_run_is_active:
        raise AgentError("There is no active test run to attach artifact")

    try:
        api = ZebrunnerAPI(zebrunner_context.settings.server.hostname, zebrunner_context.settings.server.access_token)
        api.send_artifact(path, zebrunner_context.test_run_id)
    except AgentApiError as e:
        logging.error("Failed to attach test run artifact", exc_info=e)


def attach_test_artifact_reference(name: str, ref: str) -> None:
    """
    Send artifact reference to zebrunner and attach it to test. Artifact reference is a URL
    """
    if not zebrunner_context.test_is_active:
        raise AgentError("There is no active test to attach artifact reference")

    try:
        api = ZebrunnerAPI(zebrunner_context.settings.server.hostname, zebrunner_context.settings.server.access_token)
        api.send_artifact_references(
            [ArtifactReferenceModel(name=name, value=ref)], zebrunner_context.test_run_id, zebrunner_context.test_id
        )
    except AgentApiError as e:
        logging.error("Failed to attach test artifact reference", exc_info=e)


def attach_test_run_artifact_reference(name: str, ref: str) -> None:
    """
    Send artifact reference to zebrunner and attach it to test run. Artifact reference is a URL
    """
    if not zebrunner_context.test_run_is_active:
        raise AgentError("There is no active test run to attach artifact reference")

    try:
        api = ZebrunnerAPI(zebrunner_context.settings.server.hostname, zebrunner_context.settings.server.access_token)
        api.send_artifact_references([ArtifactReferenceModel(name=name, value=ref)], zebrunner_context.test_run_id)
    except AgentApiError as e:
        logging.error("Failed to attach test run artifact reference", exc_info=e)


def attach_test_label(name: str, value: str) -> None:
    """
    Attach label to test in zebrunner
    """
    if not zebrunner_context.test_is_active:
        raise AgentError("There is no active test to attach label")

    try:
        api = ZebrunnerAPI(zebrunner_context.settings.server.hostname, zebrunner_context.settings.server.access_token)
        api.send_labels([LabelModel(key=name, value=value)], zebrunner_context.test_run_id, zebrunner_context.test_id)
    except AgentApiError as e:
        logging.error("Failed to attach label to test", exc_info=e)


def attach_test_run_label(name: str, value: str) -> None:
    """
    Attach label to test run in zebrunner
    """
    if not zebrunner_context.test_run_is_active:
        raise AgentError("There is no active test run to attach label")

    try:
        api = ZebrunnerAPI(zebrunner_context.settings.server.hostname, zebrunner_context.settings.server.access_token)
        api.send_labels([LabelModel(key=name, value=value)], zebrunner_context.test_run_id)
    except AgentApiError as e:
        logging.error("Failed to attach label to test run", exc_info=e)



import json
import logging
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Type

import dotenv
import yaml
from pydantic import BaseModel
from pydantic.utils import deep_update

PREFIX = "reporting"
logger = logging.getLogger(__name__)


class TestRunSettings(BaseModel):
    """
    A class that inherit from BaseModel and represents test_run settings.
    """

    display_name: str = "Default Suite"
    build: Optional[str] = None
    environment: Optional[str] = None
    context: Optional[str] = None
    treat_skips_as_failures: bool = True


class ServerSettings(BaseModel):
    """
    A class that inherit from BaseModel and represents server settings.
    """

    hostname: str
    access_token: str


class NotificationsSettings(BaseModel):
    """
    A class that inherit from BaseModel and represents notifications settings.
    """

    slack_channels: Optional[str] = None
    ms_teams_channels: Optional[str] = None
    emails: Optional[str] = None
    notify_on_each_failure: bool = False


class MilestoneSettings(BaseModel):
    """
    A class that inherit from BaseModel and represents milestone settings.
    """

    id: Optional[str]
    name: Optional[str]


class ZebrunnerSettings(BaseModel):
    """
    Zebrunner settings provided by launcher
    """

    @property
    def desired_capabilities(self) -> Optional[dict]:
        try:
            return json.loads(self.capabilities)  # type: ignore
        except (ValueError, TypeError) as e:
            logger.log(logging.WARN, "Failed to serialize ZEBRUNNER_CAPABILITIES option")
            logger.log(logging.WARN, e)
            return None

    capabilities: Optional[str] = None
    hub_url: Optional[str] = None


class Settings(BaseModel):
    """
    A class that inherit from BaseModel and represents some settings.
    """

    enabled: bool = True
    project_key: str = "DEF"
    send_logs: bool = True
    server: ServerSettings
    run: TestRunSettings = TestRunSettings()
    notification: Optional[NotificationsSettings] = None
    milestone: Optional[MilestoneSettings] = None
    zebrunner: Optional[ZebrunnerSettings] = None


def _list_settings(model: Type[BaseModel]) -> List:
    """
    Extracts and returns a list with all model fields. Also goes deeper into fields that extend from BaseModel and
    extract theirs fields too.

    Args:
        model (Type[BaseModel]): A model to list its fields.

    Returns:
        setting_names (list): List with all model fields.
    """
    setting_names = []
    for field_name, field_value in model.__fields__.items():
        field_list = [field_name]
        if issubclass(field_value.type_, BaseModel):
            inner_fields = _list_settings(field_value.type_)
            inner_fields = [field_list + inner for inner in inner_fields]
            setting_names += inner_fields
        else:
            setting_names.append(field_list)

    return setting_names


def _put_by_path(settings_dict: dict, path: List[str], value: Any) -> None:
    """
    Creates a dictionary with the first item in path as key and set value as its value if the amount of
    items in path is one.
    Otherwise, creates a set of nested dictionaries, with the first item in path at the top of the head.

    Args:
        settings_dict (dict): Dictionary with settings fields.
        path (List[str]): Strings to be set as dictionary keys.
        value: Some value to be set to las dictionary key.
    """
    if len(path) == 1:
        settings_dict[path[0]] = value
    else:
        current_dict = settings_dict.get(path[0], {})
        _put_by_path(current_dict, path[1:], value)
        settings_dict[path[0]] = current_dict


def _get_by_path(settings_dict: dict, path: List[str], default_value: Any = None) -> Any:
    """
    Returns the value of first path item key if path list has only one element.
    Otherwise, returns values of every key in path list recursively.

    Args:
        settings_dict (dict):
        path (List[str]):
        default_value (optional):
    """
    if len(path) == 1:
        return settings_dict.get(path[0], default_value)
    else:
        inner_dict = settings_dict.get(path[0], {})
        return _get_by_path(inner_dict, path[1:], default_value)


def _load_env(path_list: List[List[str]]) -> dict:
    """"""
    dotenv.load_dotenv(".env")
    settings: Dict[str, Any] = {}
    for path in path_list:
        env_name = "_".join(path).upper()
        env_variable = os.getenv(env_name)
        prefix_env_name = "_".join([PREFIX] + path).upper()
        prefix_env_value = os.getenv(prefix_env_name)
        if prefix_env_value is not None:
            env_variable = prefix_env_value

        if env_variable is not None:
            _put_by_path(settings, path, env_variable)

    return settings


def _load_yaml(path_list: List[List[str]]) -> Dict[str, Any]:
    settings: Dict[str, Any] = {}
    filename = Path("agent.yaml")
    if not filename.exists():
        filename = Path("agent.yml")
        if not filename.exists():
            return settings

    yaml_settings = yaml.safe_load(filename.read_text())
    for setting_path in path_list:
        yaml_path = [name.replace("_", "-") for name in setting_path]
        setting_value = _get_by_path(yaml_settings, yaml_path)

        prefix_yaml_path = [name.replace("_", "-") for name in [PREFIX] + setting_path]
        prefix_settings_value = _get_by_path(yaml_settings, prefix_yaml_path)
        if prefix_settings_value is not None:
            setting_value = prefix_settings_value

        if setting_value is not None:
            _put_by_path(settings, setting_path, setting_value)

    return settings


def load_settings() -> Settings:
    settings_path_list = _list_settings(Settings)
    settings: Dict[str, Any] = {}
    yaml_settings = _load_yaml(settings_path_list)
    settings = deep_update(settings, yaml_settings)
    env_settings = _load_env(settings_path_list)
    settings = deep_update(settings, env_settings)
    return Settings(**settings)



from pytest_zebrunner.tcm.base import BaseTcm


class Zebrunner(BaseTcm):
    _SYNC_ENABLED = "com.zebrunner.app/tcm.zebrunner.sync.enabled"
    _SYNC_REAL_TIME = "com.zebrunner.app/tcm.zebrunner.sync.real-time"
    _RUN_ID = "com.zebrunner.app/tcm.zebrunner.test-run-id"
    _CASE_ID = "com.zebrunner.app/tcm.zebrunner.test-case-key"

    @staticmethod
    def disable_sync() -> None:
        Zebrunner._verify_no_tests()
        Zebrunner._attach_label(Zebrunner._SYNC_ENABLED, "false")

    @staticmethod
    def enable_real_time_sync() -> None:
        Zebrunner._verify_no_tests()
        Zebrunner._attach_label(Zebrunner._SYNC_REAL_TIME, "true")

    @staticmethod
    def set_run_id(run_id: str) -> None:
        Zebrunner._verify_no_tests()
        Zebrunner._attach_label(Zebrunner._RUN_ID, run_id)

    @staticmethod
    def set_test_case_key(test_case_key: str) -> None:
        Zebrunner._attach_label(Zebrunner._CASE_ID, test_case_key)



import logging
from typing import Any, List, Optional, Tuple

import pytest
from _pytest._code.code import ExceptionChainRepr, ReprExceptionInfo
from _pytest.nodes import Item
from _pytest.reports import TestReport
from selenium.webdriver.common.options import BaseOptions

from pytest_zebrunner.api.client import ZebrunnerAPI
from pytest_zebrunner.api.models import (
    ArtifactReferenceModel,
    CorrelationDataModel,
    FinishTestModel,
    FinishTestSessionModel,
    MilestoneModel,
    NotificationsModel,
    NotificationsType,
    NotificationTargetModel,
    StartTestModel,
    StartTestRunModel,
    StartTestSessionModel,
    TestRunConfigModel,
    TestStatus,
)
from pytest_zebrunner.ci_loaders import CiContextLoader
from pytest_zebrunner.context import Test, TestRun, zebrunner_context
from pytest_zebrunner.errors import AgentApiError
from pytest_zebrunner.tcm.test_rail import TestRail
from pytest_zebrunner.tcm.xray import Xray
from pytest_zebrunner.tcm.zephyr import Zephyr
from pytest_zebrunner.tcm.zebrunner import Zebrunner
from pytest_zebrunner.zebrunner_logging import ZebrunnerHandler

logger = logging.getLogger(__name__)


class ReportingService:
    def __init__(self) -> None:
        self.api = ZebrunnerAPI(
            zebrunner_context.settings.server.hostname, zebrunner_context.settings.server.access_token
        )

    def authorize(self) -> None:
        if not self.api._authenticated:
            self.api.auth()

    def get_notification_configurations(self) -> Optional[NotificationsModel]:
        settings = zebrunner_context.settings
        if not settings.notification:
            return None

        targets = []
        if settings.notification.emails:
            targets.append(
                NotificationTargetModel(
                    type=NotificationsType.EMAIL_RECIPIENTS.value, value=settings.notification.emails
                )
            )
        if settings.notification.slack_channels:
            targets.append(
                NotificationTargetModel(
                    type=NotificationsType.SLACK_CHANNELS.value, value=settings.notification.slack_channels
                )
            )
        if settings.notification.ms_teams_channels:
            targets.append(
                NotificationTargetModel(
                    type=NotificationsType.MS_TEAMS_CHANNELS.value, value=settings.notification.ms_teams_channels
                )
            )
        return NotificationsModel(notify_on_each_failure=settings.notification.notify_on_each_failure, targets=targets)

    def start_test_run(self) -> None:
        self.authorize()
        settings = zebrunner_context.settings
        test_run = TestRun(settings.run.display_name, settings.run.environment, settings.run.build)
        zebrunner_context.test_run = test_run
        milestone = (
            MilestoneModel(id=settings.milestone.id, name=settings.milestone.name) if settings.milestone else None
        )

        start_run_model = StartTestRunModel(
            name=test_run.name,
            framework="pytest",
            config=TestRunConfigModel(
                environment=test_run.environment,
                build=test_run.build,
                treat_skips_as_failures=settings.run.treat_skips_as_failures,
            ),
            milestone=milestone,
            ci_context=CiContextLoader.resolve_ci_context(),
            notifications=self.get_notification_configurations(),
        )

        if settings.run.context:
            try:
                zebrunner_run_context = self.api.get_rerun_tests(settings.run.context)
            except AgentApiError as e:
                logging.error("Failed to get rerun tests", exc_info=e)
                pytest.exit("Failed to get rerun tests")

            start_run_model.uuid = zebrunner_run_context.test_run_uuid
            if not zebrunner_run_context.run_allowed:
                pytest.exit(f"Run not allowed by zebrunner! Reason: {zebrunner_run_context.reason}")
            if zebrunner_run_context.run_only_specific_tests and not zebrunner_run_context.tests_to_run:
                pytest.exit("Aborted. No tests to run!!")
        try:
            test_run.zebrunner_id = self.api.start_test_run(settings.project_key, start_run_model)
        except AgentApiError as e:
            logging.error("Failed to start test run", exc_info=e)
            pytest.exit("Failed to start test run")

    def start_test(self, report: TestReport) -> None:
        if not zebrunner_context.test_run_is_active:
            return

        self.authorize()
        test = Test(
            name=".".join(report.nodeid.split("::")[1:]),
            file=report.nodeid.split("::")[0],
            maintainers=report.maintainers,
            labels=report.labels,
        )
        zebrunner_context.test = test
        start_model = StartTestModel(
            name=test.name,
            class_name=test.file,
            method_name=test.name,
            maintainer=",".join(test.maintainers),
            labels=[{"key": label[0], "value": label[1]} for label in test.labels],
            correlation_data=CorrelationDataModel(name=test.name).json(),
        )

        test_id = self._find_attribute(report.user_properties, "zebrunner_id")
        try:
            if test_id:
                test.zebrunner_id = self.api.update_test(zebrunner_context.test_run_id, int(test_id), start_model)
            else:
                test.zebrunner_id = self.api.start_test(zebrunner_context.test_run_id, start_model)
        except AgentApiError as e:
            logging.error("Failed to start test", exc_info=e)
            return

        if report.artifact_references:
            references = [ArtifactReferenceModel(name=x[0], value=x[1]) for x in report.artifact_references]
            try:
                self.api.send_artifact_references(references, zebrunner_context.test_run_id, zebrunner_context.test_id)
            except AgentApiError as e:
                logging.error("Failed to send artifact reference", exc_info=e)

        if report.artifacts:
            for artifact in report.artifacts:
                try:
                    self.api.send_artifact(artifact, zebrunner_context.test_run_id, zebrunner_context.test_id)
                except AgentApiError as e:
                    logging.error("Failed to send artifact", exc_info=e)

    def finish_test(self, report: TestReport) -> None:
        # Test was reverted skip finishing
        if not zebrunner_context.test_is_active:
            return

        if zebrunner_context.test and zebrunner_context.test.is_reverted:
            return

        self.authorize()
        is_skip = report.when == "setup" and report.outcome == "skipped"
        is_xfail = hasattr(report, "wasxfail")

        reason = None
        if report.passed:
            status = TestStatus.PASSED
        elif is_skip or is_xfail:
            status = TestStatus.SKIPPED
        else:
            status = TestStatus.FAILED

        if is_xfail:
            reason = report.wasxfail
        elif is_skip:
            reason = report.longrepr[-1] if isinstance(report.longrepr, tuple) else str(report.longrepr)
        else:
            # Following this changelog check if it's string or ReprExceptionInfo
            # https://docs.pytest.org/en/6.2.x/changelog.html?highlight=reprexceptioninfo#pytest-6-0-0rc1-2020-07-08
            reason = report.longrepr
            if isinstance(report.longrepr, ReprExceptionInfo) or isinstance(report.longrepr, ExceptionChainRepr):
                reason = report.longrepr.reprcrash.message + "\n\n" + (str(reason) if reason else "")  # type: ignore

        test_rail_case_ids = getattr(report, "test_rail_case_ids", None)
        if test_rail_case_ids:
            for case_id in test_rail_case_ids:
                TestRail.set_case_id(case_id)

        xray_case_ids = getattr(report, "xray_case_ids", None)
        if xray_case_ids:
            for case_id in xray_case_ids:
                Xray.set_test_key(case_id)

        zephyr_case_ids = getattr(report, "zephyr_case_ids", None)
        if zephyr_case_ids:
            for case_id in zephyr_case_ids:
                Zephyr.set_test_case_key(case_id)

        zebrunner_case_keys = getattr(report, "zebrunner_case_keys", None)
        if zebrunner_case_keys:
            for case_key in zebrunner_case_keys:
                Zebrunner.set_test_case_key(case_key)

        try:
            self.api.finish_test(
                zebrunner_context.test_run_id,
                zebrunner_context.test_id,
                FinishTestModel(
                    result=status.value,
                    reason=reason,
                ),
            )
        except AgentApiError as e:
            logging.error("failed to finish test", exc_info=e)

        zebrunner_context.test = None

    def finish_test_run(self) -> None:
        if not zebrunner_context.test_run_is_active:
            return

        self.authorize()
        try:
            self.api.finish_test_run(zebrunner_context.test_run_id)
        except AgentApiError as e:
            logging.error("failed to finish test run", exc_info=e)

        loggers = [logging.getLogger(name) for name in logging.root.manager.loggerDict]  # type: ignore
        handlers: List[ZebrunnerHandler] = []
        for logger in loggers:
            handlers += [handler for handler in logger.handlers if isinstance(handler, ZebrunnerHandler)]

        for handler in handlers:
            handler.push_logs()

    def start_test_session(self, session_id: str, options: BaseOptions, test_ids: List[int]) -> Optional[str]:
        if not zebrunner_context.test_run_is_active:
            return None

        self.authorize()
        try:
            zebrunner_session_id = self.api.start_test_session(
                zebrunner_context.test_run_id,
                StartTestSessionModel(
                    session_id=session_id,
                    desired_capabilities=options.to_capabilities(),
                    capabilities=options.capabilities,
                    test_ids=test_ids,
                ),
            )
        except AgentApiError as e:
            logging.error("failed to start test session", exc_info=e)
            return None

        return zebrunner_session_id

    def add_test_to_session(self, zebrunner_session_id: str, related_tests: List[int]) -> None:
        if not zebrunner_context.test_run_is_active:
            return

        self.authorize()
        try:
            self.api.add_tests_to_session(zebrunner_context.test_run_id, zebrunner_session_id, related_tests)
        except AgentApiError as e:
            logging.error("failed to add tests to session", exc_info=e)

    def finish_test_session(self, zebrunner_session_id: str, related_tests: List[int]) -> None:
        if not zebrunner_context.test_run_is_active:
            return

        self.authorize()
        try:
            self.api.finish_test_session(
                zebrunner_context.test_run_id,
                zebrunner_session_id,
                FinishTestSessionModel(test_ids=related_tests),
            )
        except AgentApiError as e:
            logging.error("failed to finish test session", exc_info=e)

    def _find_attribute(self, properties: List[Tuple[str, Any]], name: str) -> Any:
        for property_name, value in properties:
            if property_name == name:
                return value

        return None

    def filter_test_items(self, items: List[Item]) -> List[Item]:
        run_context = zebrunner_context.settings.run.context
        if run_context is None:
            return items

        try:
            self.authorize()
            context_data = self.api.get_rerun_tests(run_context)
        except AgentApiError as e:
            logging.error("failed to filter tests", exc_info=e)
            return items

        rerun_tests = {x.name: x for x in context_data.tests_to_run if x.correlation_data is not None}

        if not context_data.tests_to_run:
            return items

        rerun_items: List[Item] = []
        for item in items:
            if item.name in rerun_tests:
                test = rerun_tests[item.name]
                item.user_properties.append(("zebrunner_id", test.id))
                rerun_items.append(item)

        return rerun_items



import logging
from typing import Optional

from pytest_zebrunner.api.models import LabelModel, PlatformModel

from .api.client import ZebrunnerAPI
from .context import zebrunner_context
from .errors import AgentApiError, AgentError


class CurrentTestRun:
    @staticmethod
    def set_build(build: str) -> None:
        if not build.strip():
            raise AgentError("Build must not be empty")

        if not zebrunner_context.test_run_is_active:
            raise AgentError("There is not active test run to set build")

        settings = zebrunner_context.settings
        try:
            api = ZebrunnerAPI(settings.server.hostname, settings.server.access_token)
            api.patch_test_run_build(zebrunner_context.test_run_id, build)
        except AgentApiError as e:
            logging.error("Failed to set build", exc_info=e)

    @staticmethod
    def set_locale(locale: str) -> None:
        if not locale.strip():
            raise AgentError("Locale must no be empty")
        if not zebrunner_context.test_run_is_active:
            raise AgentError("There is not active test run to set locale")

        label = "com.zebrunner.app/sut.locale"
        settings = zebrunner_context.settings
        try:
            api = ZebrunnerAPI(settings.server.hostname, settings.server.access_token)
            api.send_labels([LabelModel(key=label, value=locale)], zebrunner_context.test_run_id, None)
        except AgentApiError as e:
            logging.error("failed to set locale", exc_info=e)

    @staticmethod
    def set_platform(name: str) -> None:
        if not name.strip():
            raise AgentError("Platform must not be empty")

        CurrentTestRun.set_platform_version(name, None)

    @staticmethod
    def set_platform_version(name: str, version: Optional[str]) -> None:
        if not name.strip():
            raise AgentError("Platform must not be empty")
        if not zebrunner_context.test_run_is_active:
            raise AgentError("There is not active test run to set platform")

        settings = zebrunner_context.settings
        try:
            api = ZebrunnerAPI(settings.server.hostname, settings.server.access_token)
            api.set_test_run_platform(
                zebrunner_context.test_run_id,
                PlatformModel(name=name, version=version),
            )
        except AgentApiError as e:
            logging.error("Failed to set platform", exc_info=e)



import logging

from pytest_zebrunner.api.client import ZebrunnerAPI
from pytest_zebrunner.context import zebrunner_context
from pytest_zebrunner.errors import AgentApiError, AgentError


class CurrentTest:
    @staticmethod
    def revert_registration() -> None:
        if not zebrunner_context.test_is_active:
            raise AgentError("There is not active test to revert")

        settings = zebrunner_context.settings
        try:
            api = ZebrunnerAPI(settings.server.hostname, settings.server.access_token)
            api.reverse_test_registration(zebrunner_context.test_run_id, zebrunner_context.test_id)
            if zebrunner_context.test:
                zebrunner_context.test.is_reverted = True
        except AgentApiError as e:
            logging.error("Failed to revert test registration", exc_info=e)



from pytest_zebrunner.tcm.base import BaseTcm


class TestRail(BaseTcm):
    _SYNC_ENABLED = "com.zebrunner.app/tcm.testrail.sync.enabled"
    _SYNC_REAL_TIME = "com.zebrunner.app/tcm.testrail.sync.real-time"
    _INCLUDE_ALL = "com.zebrunner.app/tcm.testrail.include-all-cases"
    _SUITE_ID = "com.zebrunner.app/tcm.testrail.suite-id"
    _RUN_ID = "com.zebrunner.app/tcm.testrail.run-id"
    _RUN_NAME = "com.zebrunner.app/tcm.testrail.run-name"
    _MILESTONE = "com.zebrunner.app/tcm.testrail.milestone"
    _ASSIGNEE = "com.zebrunner.app/tcm.testrail.assignee"
    _CASE_ID = "com.zebrunner.app/tcm.testrail.case-id"

    @staticmethod
    def disable_sync() -> None:
        TestRail._verify_no_tests()
        TestRail._attach_label(TestRail._SYNC_ENABLED, "false")

    @staticmethod
    def enable_real_time_sync() -> None:
        TestRail._verify_no_tests()
        TestRail._attach_label(TestRail._SYNC_REAL_TIME, "true")
        TestRail._attach_label(TestRail._INCLUDE_ALL, "true")

    @staticmethod
    def include_all_test_cases_in_new_run() -> None:
        TestRail._verify_no_tests()
        TestRail._attach_label(TestRail._INCLUDE_ALL, "true")

    @staticmethod
    def set_suite_id(suite_id: str) -> None:
        TestRail._verify_no_tests()
        TestRail._attach_label(TestRail._SUITE_ID, suite_id)

    @staticmethod
    def set_run_id(run_id: str) -> None:
        TestRail._verify_no_tests()
        TestRail._attach_label(TestRail._RUN_ID, run_id)

    @staticmethod
    def set_run_name(run_name: str) -> None:
        TestRail._verify_no_tests()
        TestRail._attach_label(TestRail._RUN_NAME, run_name)

    @staticmethod
    def set_milestone(milestone: str) -> None:
        TestRail._verify_no_tests()
        TestRail._attach_label(TestRail._MILESTONE, milestone)

    @staticmethod
    def set_assignee(assignee: str) -> None:
        TestRail._verify_no_tests()
        TestRail._attach_label(TestRail._ASSIGNEE, assignee)

    @staticmethod
    def set_case_id(test_case_id: str) -> None:
        TestRail._attach_label(TestRail._CASE_ID, test_case_id)



from pytest_zebrunner.tcm.base import BaseTcm


class Xray(BaseTcm):
    _SYNC_ENABLED = "com.zebrunner.app/tcm.xray.sync.enabled"
    _SYNC_REAL_TIME = "com.zebrunner.app/tcm.xray.sync.real-time"
    _EXECUTION_KEY = "com.zebrunner.app/tcm.xray.test-execution-key"
    _TEST_KEY = "com.zebrunner.app/tcm.xray.test-key"

    @staticmethod
    def disable_sync() -> None:
        Xray._verify_no_tests()
        Xray._attach_label(Xray._SYNC_ENABLED, "false")

    @staticmethod
    def enable_real_time_sync() -> None:
        Xray._verify_no_tests()
        Xray._attach_label(Xray._SYNC_REAL_TIME, "true")

    @staticmethod
    def set_execution_key(execution_key: str) -> None:
        Xray._verify_no_tests()
        Xray._attach_label(Xray._EXECUTION_KEY, execution_key)

    @staticmethod
    def set_test_key(test_key: str) -> None:
        Xray._attach_label(Xray._TEST_KEY, test_key)



class Singleton(type):
    """
    A class that inherit form 'type' and allows to implement Singleton Pattern.
    """

    __instance = None

    def __call__(cls, *args, **kwargs):  # type: ignore
        if not isinstance(cls.__instance, cls):
            cls.__instance = super(Singleton, cls).__call__(*args, **kwargs)

        return cls.__instance



"""Test case management integration package"""

from .test_rail import TestRail  # noqa: F401
from .xray import Xray  # noqa: F401
from .zephyr import Zephyr  # noqa: F401
from .zebrunner import Zebrunner  # noqa: F401



import functools
import json
import logging
import time
from datetime import datetime, timezone
from pathlib import Path
from pprint import pformat
from typing import List, Optional, Union

import httpx
from httpx import Client, Request, Response

from pytest_zebrunner.api.models import (
    ArtifactReferenceModel,
    AttachTestsToSessionModel,
    FinishTestModel,
    FinishTestSessionModel,
    LabelModel,
    LogRecordModel,
    PlatformModel,
    RerunDataModel,
    StartTestModel,
    StartTestRunModel,
    StartTestSessionModel,
)
from pytest_zebrunner.errors import AgentApiError
from pytest_zebrunner.utils import Singleton

logger = logging.getLogger(__name__)


def log_response(response: Response, log_level: int = logging.DEBUG) -> None:
    """Logger customized configuration"""
    request = response.request
    request.read()

    logger.log(
        log_level,
        f"Request {request.method} {request.url}\n"
        f"Headers: \n{pformat(dict(request.headers))}\n\n"
        f"Content: \n{request.content!r}\n\n"
        f"Response Code: {response.status_code}\n"
        f" Content: \n{pformat(response.json())}",
    )


def retry(method, times: int = 6, first_wait: float = 5):
    """
    Wrapper for retry to connect and wait between retries.
    """

    def wrapper(*args, **kwargs):
        for x in range(times):
            try:
                result = method(*args, **kwargs)
            except AgentApiError as e:
                logger.error(e)
                if x != times:
                    logger.info(f"Retry number: {x + 1}")
                    time.sleep(first_wait * (2**x))
            else:
                return result

    return wrapper


class ZebrunnerAPI(metaclass=Singleton):
    """
    A singleton Zebrunner API representation
    """

    _authenticated = False

    def __init__(self, service_url: str = None, access_token: str = None):
        if service_url and access_token:
            self.service_url = service_url.rstrip("/")
            self.access_token = access_token
            self._client = Client(timeout=60.0)
            self._auth_token = None
            self._authenticated = False

    def _sign_request(self, request: Request) -> Request:
        """
        Returns a request with the _auth_token set to the authorization request header.
        """
        request.headers["Authorization"] = f"Bearer {self._auth_token}"
        return request

    @retry
    def auth(self) -> None:
        """
        Validates the user access token with http post method and if it is correct, authenticates the user.
        """
        if not self.access_token or not self.service_url:
            return

        url = f"{self.service_url}/api/iam/v1/auth/refresh"
        try:
            response = self._client.post(url, json={"refreshToken": self.access_token})
        except httpx.RequestError as e:
            raise AgentApiError("Failed to authorize zebrunner agent", e)

        if not response.is_success:
            raise AgentApiError(
                "Failed to authorize zebrunner agent", {"status_code": response.status_code, "body": response.json()}
            )

        self._auth_token = response.json()["authToken"]
        self._client.auth = self._sign_request  # type: ignore
        self._authenticated = True

    def start_test_run(self, project_key: str, body: StartTestRunModel) -> Optional[int]:
        """
        Send POST request creating new test run. Raise ApiAgentException if request failed
        """
        url = f"{self.service_url}/api/reporting/v1/test-runs"

        try:
            response = self._client.post(
                url, params={"projectKey": project_key}, json=body.dict(exclude_none=True, by_alias=True)
            )
        except httpx.RequestError as e:
            raise AgentApiError("Failed to create test run", e)

        if not response.is_success:
            raise AgentApiError(
                "Failed to create test run. Non successful response code",
                {"status_code": response.status_code, "body": response.json()},
            )

        return response.json()["id"]

    def start_test(self, test_run_id: int, body: StartTestModel) -> Optional[int]:
        """
        Send POST request creating new test. Raise AgentApiError in case of any exceptions
        """
        url = f"{self.service_url}/api/reporting/v1/test-runs/{test_run_id}/tests"

        try:
            response = self._client.post(url, json=body.dict(exclude_none=True, by_alias=True))
        except httpx.RequestError as e:
            raise AgentApiError("Failed to create test", e)

        if not response.is_success:
            raise AgentApiError(
                "Failed to create test. Non successful response code",
                {"status_code": response.status_code, "body": response.json()},
            )

        return response.json()["id"]

    def update_test(self, test_run_id: int, test_id: int, test: StartTestModel) -> Optional[int]:
        """
        Send PUT request updating some test. Raise AgentApiError in case of any exceptions
        """

        url = f"{self.service_url}/api/reporting/v1/test-runs/{test_run_id}/tests/{test_id}/"
        try:
            response = self._client.post(url, json=test.dict(exclude_none=True, by_alias=True))
        except httpx.RequestError as e:
            raise AgentApiError("Failed to update test", e)

        if not response.is_success:
            raise AgentApiError(
                "Failed to update test. Non successful status code",
                {"status_code": response.status_code, "body": response.json()},
            )

        return response.json()["id"]

    def finish_test(self, test_run_id: int, test_id: int, body: FinishTestModel) -> None:
        """
        Send PUT request finishing current test. Raise AgentApiError in case of any exceptions
        """
        url = f"{self.service_url}/api/reporting/v1/test-runs/{test_run_id}/tests/{test_id}"

        try:
            response = self._client.put(url, json=body.dict(exclude_none=True, by_alias=True))
        except httpx.RequestError as e:
            raise AgentApiError("Failed to finish test", e)

        if not response.is_success:
            raise AgentApiError(
                "Failed to finish test. Non  successful status code",
                {"status_code": response.status_code, "body": response.json()},
            )

    def finish_test_run(self, test_run_id: int) -> None:
        """
        Send PUT request finishing current test run. Raise AgentApiError in case of any exceptions
        """
        url = f"{self.service_url}/api/reporting/v1/test-runs/{test_run_id}"
        try:
            response = self._client.put(
                url,
                json={"endedAt": (datetime.utcnow().replace(tzinfo=timezone.utc).isoformat())},
            )
        except httpx.RequestError as e:
            raise AgentApiError("Failed to finish test run", e)

        if not response.is_success:
            raise AgentApiError(
                "Failed to finish test run. Non  successful status code",
                {"status_code": response.status_code, "body": response.json()},
            )

    def send_logs(self, test_run_id: int, logs: List[LogRecordModel]) -> None:
        """
        Send POST request uploading logs. Raise AgentApiError in case of any exceptions
        """
        url = f"{self.service_url}/api/reporting/v1/test-runs/{test_run_id}/logs"

        body = [x.dict(exclude_none=True, by_alias=True) for x in logs]
        try:
            response = self._client.post(url, json=body)
        except httpx.RequestError as e:
            raise AgentApiError("Failed to send logs", e)

        if not response.is_success:
            raise AgentApiError(
                "Failed to send logs. Non successful status code",
                {"status_code": response.status_code, "body": response.json()},
            )

    def send_screenshot(self, test_run_id: int, test_id: int, image_path: Union[str, Path]) -> None:
        """
        Send screenshot to zebrunner. Raise AgentApiError in case of any exceptions
        """
        url = f"{self.service_url}/api/reporting/v1/test-runs/{test_run_id}/tests/{test_id}/screenshots"
        with open(image_path, "rb") as image:
            try:
                response = self._client.post(
                    url,
                    content=image.read(),
                    headers={
                        "Content-Type": "image/png",
                        "x-zbr-screenshot-captured-at": str(round(time.time() * 1000)),
                    },
                )
            except httpx.RequestError as e:
                raise AgentApiError("Failed to send screenshot", e)

            if not response.is_success:
                raise AgentApiError(
                    "Failed to send screenshot. Non successful status code",
                    {"status_code": response.status_code, "body": response.json()},
                )

    def send_artifact(self, filename: Union[str, Path], test_run_id: int, test_id: Optional[int] = None) -> None:
        """
        Send artifact to zebrunner. Attach it to test run if test_id is None else attach it to test.
        Raise AgentApiError in case of any exceptions
        """
        if test_id:
            url = f"{self.service_url}/api/reporting/v1/test-runs/{test_run_id}/tests/{test_id}/artifacts"
        else:
            url = f"{self.service_url}/api/reporting/v1/test-runs/{test_run_id}/artifacts"

        with open(filename, "rb") as file:
            try:
                response = self._client.post(url, files={"file": file})
            except httpx.RequestError as e:
                raise AgentApiError("Failed to send artifact", e)

            if not response.is_success:
                raise AgentApiError(
                    "Failed to send artifact. Non successful status code",
                    {"status_code": response.status_code, "body": response.json()},
                )

    def send_artifact_references(
        self, references: List[ArtifactReferenceModel], test_run_id: int, test_id: Optional[int] = None
    ) -> None:
        """
        Send artifact reference to zebrunner. Attach it to test run if test_id is None else attach it to test.
        Raise AgentApiError in case of any exceptions
        """
        if test_id:
            url = f"{self.service_url}/api/reporting/v1/test-runs/{test_run_id}/tests/{test_id}/artifact-references"
        else:
            url = f"{self.service_url}/api/reporting/v1/test-runs/{test_run_id}/artifact-references/"
        json_items = [item.dict(exclude_none=True, by_alias=True) for item in references]

        try:
            response = self._client.put(url, json={"items": json_items})
        except httpx.RequestError as e:
            raise AgentApiError("Failed to send artifact reference", e)

        if not response.is_success:
            raise AgentApiError(
                "Failed to send artifact reference. Non successful status code",
                {"status_code": response.status_code, "body": response.json()},
            )

    def send_labels(self, labels: List[LabelModel], test_run_id: int, test_id: Optional[int] = None) -> None:
        """
        Send labels to zebrunner. Attach it to test run if test_id is None else attach it to test.
        Raise AgentApiError in case of any exceptions
        """
        if test_id:
            url = f"{self.service_url}/api/reporting/v1/test-runs/{test_run_id}/tests/{test_id}/labels"
        else:
            url = f"{self.service_url}/api/reporting/v1/test-runs/{test_run_id}/labels"
        labels_json = [label.dict(exclude_none=True, by_alias=True) for label in labels]

        try:
            response = self._client.put(url, json={"items": labels_json})
        except httpx.RequestError as e:
            raise AgentApiError("Failed to send labels", e)

        if not response.is_success:
            raise AgentApiError(
                "Failed to send labels. Non successful status code",
                {"status_code": response.status_code, "body": response.json()},
            )

    def start_test_session(self, test_run_id: int, body: StartTestSessionModel) -> Optional[str]:
        """
        Send POST request starting test session. Raise AgentApiError in case of any exceptions
        """
        url = f"{self.service_url}/api/reporting/v1/test-runs/{test_run_id}/test-sessions"

        try:
            response = self._client.post(url, json=body.dict(exclude_none=True, by_alias=True))
        except httpx.RequestError as e:
            raise AgentApiError("Failed to start session", e)

        if not response.status_code == 200:
            raise AgentApiError(
                "Failed to start session. Non successful status code",
                {"status_code": response.status_code, "body": response.json()},
            )

        return response.json().get("id")

    def add_tests_to_session(self, test_run_id: int, session_id: str, related_tests: List[int]) -> None:
        """
        Send PUT request attaching new test to test session. Raise AgentApiError in case of any exceptions
        """
        url = f"{self.service_url}/api/reporting/v1/test-runs/{test_run_id}/test-sessions/{session_id}"
        body = AttachTestsToSessionModel(test_ids=related_tests)
        try:
            response = self._client.put(url, json=body.dict(exclude_none=True, by_alias=True))
        except httpx.RequestError as e:
            raise AgentApiError("Failed to attach tests to session", e)

        if not response.is_success:
            raise AgentApiError(
                "Failed to attach tests to session. Non successful status code",
                {"status_code": response.status_code, "body": response.json()},
            )

    def finish_test_session(self, test_run_id: int, test_id: str, body: FinishTestSessionModel) -> None:
        """
        Send PUT request finishing test session. Raise AgentApiError in case of any exceptions
        """
        url = f"{self.service_url}/api/reporting/v1/test-runs/{test_run_id}/test-sessions/{test_id}"
        try:
            response = self._client.put(url, json=body.dict(exclude_none=True, by_alias=True))
        except httpx.RequestError as e:
            raise AgentApiError("Failed to start session", e)

        if not response.is_success:
            raise AgentApiError(
                "Failed to finish session. Non successful status code",
                {"status_code": response.status_code, "body": response.json()},
            )

    @retry
    def get_rerun_tests(self, run_context: str) -> RerunDataModel:
        """Exchange run context on tests to run. Raise AgentApiError in case of any exceptions"""
        url = f"{self.service_url}/api/reporting/v1/run-context-exchanges"
        run_context_dict = json.loads(run_context)
        try:
            response = self._client.post(url, json=run_context_dict)
        except httpx.RequestError as e:
            raise AgentApiError("Failed to get rerun tests", e)

        if not response.is_success:
            raise AgentApiError(
                "Failed to get rerun tests. Non successful status code",
                {"status_code": response.status_code, "body": response.json()},
            )

        response_data = response.json()
        for test in response_data["testsToRun"]:
            test["correlationData"] = json.loads(test["correlationData"]) if test["correlationData"] else None

        return RerunDataModel(**response_data)

    def reverse_test_registration(self, test_run_id: int, test_id: int) -> None:
        """Send PUT request reversing test registration. Raise AgentApiError in case of any exceptions"""

        url = f"{self.service_url}/api/reporting/v1/test-runs/{test_run_id}/tests/{test_id}"

        try:
            response = self._client.delete(url)
        except httpx.RequestError as e:
            raise AgentApiError("Failed to revert test registration", e)

        if not response.is_success:
            raise AgentApiError(
                "Failed to revert test registration. Non successful status code",
                {"status_code": response.status_code, "body": response.json()},
            )

    def set_test_run_platform(self, run_id: int, platform: PlatformModel) -> None:
        """Update test run platform. Raise AgentApiError in case of any exceptions"""
        url = f"{self.service_url}/api/reporting/v1/test-runs/{run_id}/platform"

        try:
            response = self._client.put(url, json=platform.dict(exclude_none=True, by_alias=True))
        except httpx.RequestError as e:
            raise AgentApiError("failed to set test run platform", e)

        if not response.is_success:
            raise AgentApiError(
                "Failed to set test run platform. Non successful status code",
                {"status_code": response.status_code, "body": response.json()},
            )

    def patch_test_run_build(self, run_id: int, build: str) -> None:
        """Set test run build. Raise AgentApiError in case of any exceptions"""
        url = f"{self.service_url}/api/reporting/v1/test-runs/{run_id}"

        body = {
            "op": "replace",
            "path": "/config/build",
            "value": build,
        }

        try:
            response = self._client.patch(url, json=[body])
        except httpx.RequestError as e:
            raise AgentApiError("failed to patch test run build", e)

        if not response.is_success:
            raise AgentApiError(
                "Failed to patch test run platform. Non successful status code",
                {"status_code": response.status_code, "body": response.json()},
            )

    def close(self) -> None:
        """
        Close the connection pool without block-usage.
        """
        self._client.close()



import copy
import logging
from typing import Dict, List

from selenium.webdriver.common.options import BaseOptions

from pytest_zebrunner.context import zebrunner_context
from pytest_zebrunner.reporting_service import ReportingService

logger = logging.getLogger(__name__)

# List of allowed WebDriver capabilities
# https://www.w3.org/TR/webdriver1/#capabilities
allowed_capabilities: List = [
    "browserName",
    "browserVersion",
    "platformName",
    "acceptInsecureCerts",
    "pageLoadStrategy",
    "proxy",
    "setWindowRect",
    "timeouts",
    "unhandledPromptBehavior",
]


class SeleniumSession:
    zebrunner_id: str
    id: str
    tests: List[int]

    def __init__(self, id: str) -> None:
        self.id = id
        self.tests = []


class SeleniumSessionManager:
    def __init__(self, reporting_service: ReportingService) -> None:
        self._active_sessions: Dict[str, SeleniumSession] = {}
        self._reporting_service = reporting_service

    def start_session(self, session_id: str, options: BaseOptions) -> None:
        session = SeleniumSession(session_id)
        self._active_sessions[session_id] = session
        if zebrunner_context.test_is_active:
            session.tests.append(zebrunner_context.test_id)

        zebrunner_session_id = self._reporting_service.start_test_session(
            session_id, options, self._active_sessions[session_id].tests
        )
        if zebrunner_session_id:
            session.zebrunner_id = zebrunner_session_id

    def finish_session(self, session_id: str) -> None:
        session = self._active_sessions[session_id]
        self._reporting_service.finish_test_session(session.zebrunner_id, session.tests)
        del self._active_sessions[session_id]

    def finish_all_sessions(self) -> None:
        for session_id in list(self._active_sessions.keys()):
            self.finish_session(session_id)

    def add_test(self, test_id: int) -> None:
        for session in self._active_sessions.values():
            session.tests.append(test_id)
            self._reporting_service.add_test_to_session(session.zebrunner_id, [test_id])


def inject_driver(session_manager: SeleniumSessionManager) -> None:
    try:
        from selenium.webdriver.remote.webdriver import WebDriver

        base_init = WebDriver.__init__
        base_quit = WebDriver.quit

        def init(  # type: ignore
            session: WebDriver,
            command_executor: str = "http://127.0.0.1:4444/wd/hub",
            keep_alive: bool = False,
            file_detector=None,
            options=None,
        ) -> None:  # type: ignore
            # Override capabilities and command_executor with new ones provided by zebrunner.
            opts = copy.deepcopy(options)
            if zebrunner_context.settings.zebrunner:
                zeb_settings = zebrunner_context.settings.zebrunner
                if zeb_settings.hub_url:
                    command_executor = zeb_settings.hub_url
                zeb_desired_capabilities = zeb_settings.desired_capabilities
                if zeb_desired_capabilities and opts:
                    for capability_name, capability_value in zeb_desired_capabilities.items():
                        if (capability_name in allowed_capabilities) or (":" in capability_name):
                            opts.set_capability(capability_name, capability_value)
                        else:
                            logger.warning(f"Capability '{capability_name}':'{capability_value}' was not set")

            base_init(
                self=session,
                command_executor=command_executor,
                keep_alive=keep_alive,
                file_detector=file_detector,
                options=opts,
            )
            session_manager.start_session(session.session_id, options)
            if zebrunner_context.test_is_active:
                session_manager.add_test(zebrunner_context.test_id)

        def quit(session) -> None:  # type: ignore
            session_manager.finish_session(session.session_id)
            base_quit(session)

        WebDriver.__init__ = init
        WebDriver.quit = quit

    except ImportError:
        logger.warning("Selenium library is not installed.")



import os
from enum import Enum
from typing import Dict, List, Optional, Type

from pytest_zebrunner.api.models import CiContextModel


class CiType(Enum):
    JENKINS = "JENKINS"
    TEAM_CITY = "TEAM_CITY"
    CIRCLE_CI = "CIRCLE_CI"
    TRAVIS_CI = "TRAVIS_CI"
    BAMBOO = "BAMBOO"


class CiVariables:
    CI_ENV_VARIABLE: str
    CI_TYPE: CiType
    ENV_VARIABLE_PREFIXES: List[str]


class Jenkins(CiVariables):
    """
    Jenkins env variables
    """

    CI_ENV_VARIABLE = "JENKINS_URL"
    CI_TYPE = CiType.JENKINS
    ENV_VARIABLE_PREFIXES = [
        "CVS_",
        "SVN_",
        "GIT_",
        "NODE_",
        "EXECUTOR_NUMBER",
        "JENKINS_",
        "JOB_",
        "BUILD_",
        "ROOT_BUILD_",
        "RUN_",
        "WORKSPACE",
    ]


class TeamCity(CiVariables):
    """
    TeamCity environemt variables
    """

    CI_ENV_VARIABLE = "TEAMCITY_VERSION"
    CI_TYPE = CiType.TEAM_CITY
    ENV_VARIABLE_PREFIXES = [
        "BUILD_",
        "HOSTNAME",
        "SERVER_URL",
        "TEAMCITY_",
    ]


class CircleCi(CiVariables):
    """
    CircleCI environemt variables
    """

    CI_ENV_VARIABLE = "CIRCLECI"
    CI_TYPE = CiType.CIRCLE_CI
    ENV_VARIABLE_PREFIXES = ["CIRCLE", "HOSTNAME"]


class TravisCi(CiVariables):
    """
    CircleCI environemt variables
    """

    CI_ENV_VARIABLE = "TRAVIS"
    CI_TYPE = CiType.TRAVIS_CI
    ENV_VARIABLE_PREFIXES = ["TRAVIS", "USER"]


class CiContextLoader:
    @staticmethod
    def load_context_variables(prefixes: List[str]) -> Dict[str, str]:
        """
        Returns a dictionary with environment variables that start with one of prefixes
        """
        env_variable_names = list(filter(lambda name: any([name.startswith(x) for x in prefixes]), os.environ))
        return {key: os.environ[key] for key in env_variable_names}

    @staticmethod
    def resolve_ci_context() -> Optional[CiContextModel]:
        """
        Go through list of ci variables definitions and try to load variables related to this definition.
        Return first found ci context
        """

        ci_services_variables: List[Type[CiVariables]] = [
            Jenkins,
            TeamCity,
            CircleCi,
            TravisCi,
        ]

        for ci_variables in ci_services_variables:
            if ci_variables.CI_ENV_VARIABLE not in os.environ:
                continue

            return CiContextModel(
                ci_type=ci_variables.CI_TYPE.value,
                env_variables=CiContextLoader.load_context_variables(ci_variables.ENV_VARIABLE_PREFIXES),
            )

        return None
